@@ -41,7 +41,7 @@
 
 #define MODE_TABLE_OFFSET_845G 617
 
-#define VERSION "0.5.2"
+#define VERSION "0.5.3"
 
 #define ATI_SIGNATURE1 "ATI MOBILITY RADEON"
 #define ATI_SIGNATURE2 "ATI Technologies Inc"
@@ -55,10 +55,14 @@ typedef unsigned char boolean;
 typedef unsigned int cardinal;
 
 typedef enum {
-    CT_UNKWN, CT_845G, CT_855GM, CT_865G, CT_915G, CT_915GM, CT_945G, CT_945GM 
+    CT_UNKWN, CT_830, CT_845G, CT_855GM, CT_865G, CT_915G, CT_915GM, CT_945G, CT_945GM,
+    CT_946GZ, CT_G965, CT_Q965
 } chipset_type;
 
-char * chipset_type_names[] = {"UNKNOWN", "845G", "855GM", "865G", "915G", "915GM", "945G", "945GM"};
+char * chipset_type_names[] = {
+    "UNKNOWN", "830",  "845G", "855GM", "865G", "915G", "915GM", "945G", "945GM",
+    "946GZ",   "G965", "Q965"
+};
 
 typedef enum {
     BT_UNKWN, BT_1, BT_2, BT_3
@@ -104,7 +108,9 @@ typedef struct {
 } __attribute__((packed)) vbios_modeline_type2;
 
 typedef struct {
-    byte unknown[6];
+    byte xchars;
+    byte ychars;
+    byte unknown[4];
 
     vbios_modeline_type2 modelines[];
 } __attribute__((packed)) vbios_resolution_type2;
@@ -140,8 +146,6 @@ typedef struct {
 
 
 typedef struct {
-    boolean file;
-
     cardinal chipset_id;
     chipset_type chipset;
     bios_type bios;
@@ -177,6 +181,10 @@ chipset_type get_chipset(cardinal id) {
     chipset_type type;
     
     switch (id) {
+    case 0x35758086:
+        type = CT_830;
+        break;
+
     case 0x25608086:
         type = CT_845G;
         break;
@@ -205,6 +213,18 @@ chipset_type get_chipset(cardinal id) {
         type = CT_945GM;
         break;
 
+    case 0x29708086:
+        type = CT_946GZ;
+        break;
+
+    case 0x29a08086:
+	type = CT_G965;
+	break;
+
+    case 0x29908086:
+        type = CT_Q965;
+        break;
+
     default:
         type = CT_UNKWN;
         break;
@@ -467,45 +487,39 @@ void close_vbios(vbios_map * map) {
 
 void unlock_vbios(vbios_map * map) {
 
-    if (!map->file) {
-        assert(!map->unlocked);
+    assert(!map->unlocked);
         
-        map->unlocked = TRUE;
-        
-        switch (map->chipset) {
-        case CT_UNKWN:
-            break;
-        case CT_855GM:
-            outl(0x8000005a, 0xcf8);
-            map->b1 = inb(0xcfe);
-            
-            outl(0x8000005a, 0xcf8);
-            outb(0x33, 0xcfe);
-            break;
-        case CT_845G:
-        case CT_865G:
-	case CT_915G:
-        case CT_915GM:
-        case CT_945G:
-        case CT_945GM:
-            outl(0x80000090, 0xcf8);
-            map->b1 = inb(0xcfd);
-            map->b2 = inb(0xcfe);
-            
-            outl(0x80000090, 0xcf8);
-            outb(0x33, 0xcfd);
-            outb(0x33, 0xcfe);
-            break;
-        }
-    }
-
-#if DEBUG
-    {
-        cardinal t = inl(0xcfc);
-        printf("unlock PAM: (0x%08x)\n", t);
-    }
-#endif
+    map->unlocked = TRUE;
     
+    switch (map->chipset) {
+    case CT_UNKWN:
+        break;
+    case CT_830:
+    case CT_855GM:
+        outl(0x8000005a, 0xcf8);
+        map->b1 = inb(0xcfe);
+        
+        outl(0x8000005a, 0xcf8);
+        outb(0x33, 0xcfe);
+        break;
+    case CT_845G:
+    case CT_865G:
+    case CT_915G:
+    case CT_915GM:
+    case CT_945G:
+    case CT_945GM:
+    case CT_946GZ:
+    case CT_G965:
+    case CT_Q965:
+        outl(0x80000090, 0xcf8);
+        map->b1 = inb(0xcfd);
+        map->b2 = inb(0xcfe);
+        
+        outl(0x80000090, 0xcf8);
+        outb(0x33, 0xcfd);
+        outb(0x33, 0xcfe);
+        break;
+    }
 
 #if DEBUG
     {
@@ -513,42 +527,36 @@ void unlock_vbios(vbios_map * map) {
         printf("unlock PAM: (0x%08x)\n", t);
     }
 #endif
-
 }
 
 void relock_vbios(vbios_map * map) {
 
-    if (!map->file) {
-        assert(map->unlocked);
-        map->unlocked = FALSE;
-        
-        switch (map->chipset) {
-        case CT_UNKWN:
-            break;
-        case CT_855GM:
-            outl(0x8000005a, 0xcf8);
-            outb(map->b1, 0xcfe);
-            break;
-        case CT_845G:
-        case CT_865G:
-	case CT_915G:
-        case CT_915GM:
-        case CT_945G:
-        case CT_945GM:
-            outl(0x80000090, 0xcf8);
-            outb(map->b1, 0xcfd);
-            outb(map->b2, 0xcfe);
-            break;
-        }
+    assert(map->unlocked);
+    map->unlocked = FALSE;
+    
+    switch (map->chipset) {
+    case CT_UNKWN:
+        break;
+    case CT_830:
+    case CT_855GM:
+        outl(0x8000005a, 0xcf8);
+        outb(map->b1, 0xcfe);
+        break;
+    case CT_845G:
+    case CT_865G:
+    case CT_915G:
+    case CT_915GM:
+    case CT_945G:
+    case CT_945GM:
+    case CT_946GZ:
+    case CT_G965:
+    case CT_Q965:
+        outl(0x80000090, 0xcf8);
+        outb(map->b1, 0xcfd);
+        outb(map->b2, 0xcfe);
+        break;
     }
 
-#if DEBUG
-    {
-        cardinal t = inl(0xcfc);
-        printf("relock PAM: (0x%08x)\n", t);
-    }
-#endif
-
 #if DEBUG
     {
         cardinal t = inl(0xcfc);
@@ -611,7 +619,29 @@ void list_modes(vbios_map *map, cardinal raw) {
     }
 }
 
+static void gtf_timings(int x, int y, int freq,
+        unsigned long *clock,
+        word *hsyncstart, word *hsyncend, word *hblank,
+        word *vsyncstart, word *vsyncend, word *vblank)
+{
+    int hbl, vbl, vfreq;
+
+    vbl = y + (y+1)/(20000.0/(11*freq) - 1) + 1.5;
+    vfreq = vbl * freq;
+    hbl = 16 * (int)(x * (30.0 - 300000.0 / vfreq) /
+            (70.0 + 300000.0 / vfreq) / 16.0 + 0.5);
+
+    *vsyncstart = y;
+    *vsyncend = y + 3;
+    *vblank = vbl - 1;
+    *hsyncstart = x + hbl / 2 - (x + hbl + 50) / 100 * 8 - 1;
+    *hsyncend = x + hbl / 2 - 1;
+    *hblank = x + hbl - 1;
+    *clock = (x + hbl) * vfreq / 1000;
+}
+
 void set_mode(vbios_map * map, cardinal mode, cardinal x, cardinal y, cardinal bp, cardinal htotal, cardinal vtotal) {
+    int xprev, yprev;
     cardinal i, j;
 
     for (i=0; i < map->mode_table_size; i++) {
@@ -641,29 +671,32 @@ void set_mode(vbios_map * map, cardinal mode, cardinal x, cardinal y, cardinal b
                 {
                     vbios_resolution_type2 * res = map_type2_resolution(map, map->mode_table[i].resolution);
 
+                    res->xchars = x / 8;
+                    res->ychars = y / 16 - 1;
+                    xprev = res->modelines[0].x1;
+                    yprev = res->modelines[0].y1;
+
                     for(j=0; j < 3; j++) {
                         vbios_modeline_type2 * modeline = &res->modelines[j];
                         
-                        if (modeline->x1 == res->modelines[0].x1) {
+                        if (modeline->x1 == xprev && modeline->y1 == yprev) {
                             modeline->x1 = modeline->x2 = x-1;
                             modeline->y1 = modeline->y2 = y-1;
-			    if (htotal)
-                        	modeline->htotal = htotal;
 
-			    if (vtotal)
-                        	modeline->vtotal = vtotal;
-                            
-                            /* WARNING: I don't know at all what are the formulas
-                               to calculate these values !
-                               The following lines are only an example.
-                            */
-                            modeline->clock = freqs[j] * x;
-                            
-                            modeline->hsyncstart = x+10;
-                            modeline->hsyncend   = x+20;
-                            
-                            modeline->vsyncstart = y+10;
-                            modeline->vsyncend   = y+20;
+                            gtf_timings(x, y, freqs[j], &modeline->clock,
+                                    &modeline->hsyncstart, &modeline->hsyncend,
+                                    &modeline->hblank, &modeline->vsyncstart,
+                                    &modeline->vsyncend, &modeline->vblank);
+
+                            if (htotal)
+                                modeline->htotal = htotal;
+                            else
+                                modeline->htotal = modeline->hblank;
+
+                            if (vtotal)
+                                modeline->vtotal = vtotal;
+                            else
+                                modeline->vtotal = modeline->vblank;
                         }
                     }
                 }
@@ -672,34 +705,31 @@ void set_mode(vbios_map * map, cardinal mode, cardinal x, cardinal y, cardinal b
                 {
                     vbios_resolution_type3 * res = map_type3_resolution(map, map->mode_table[i].resolution);
                     
+                    xprev = res->modelines[0].x1;
+                    yprev = res->modelines[0].y1;
+
                     for (j=0; j < 3; j++) {
                         vbios_modeline_type3 * modeline = &res->modelines[j];
                         
-                        if (modeline->x1 == res->modelines[0].x1) {
+                        if (modeline->x1 == xprev && modeline->y1 == yprev) {
                             modeline->x1 = modeline->x2 = x-1;
                             modeline->y1 = modeline->y2 = y-1;
                             
-			    if (htotal)
-                        	modeline->htotal = htotal;
+                            gtf_timings(x, y, freqs[j], &modeline->clock,
+                                    &modeline->hsyncstart, &modeline->hsyncend,
+                                    &modeline->hblank, &modeline->vsyncstart,
+                                    &modeline->vsyncend, &modeline->vblank);
+                            if (htotal)
+                                modeline->htotal = htotal;
+                            else
+                                modeline->htotal = modeline->hblank;
+                            if (vtotal)
+                                modeline->vtotal = vtotal;
+                            else
+                                modeline->vtotal = modeline->vblank;
 
-			    if (vtotal)
-                        	modeline->vtotal = vtotal;
-
-                            /* WARNING: I don't know at all what are the formulas
-                               to calculate these values !
-                               The following lines are only an example.
-                            */
-
-                            modeline->clock = freqs[i] * x;
-                            
-                            modeline->hsyncstart = x+10;
-                            modeline->hsyncend   = x+20;
-                            
-                            modeline->vsyncstart = y+10;
-                            modeline->vsyncend   = y+20;
-                            
-                            modeline->timing_h   = 0x2ff;
-                            modeline->timing_v   = 0x4ff;
+                            modeline->timing_h   = y-1;
+                            modeline->timing_v   = x-1;
                         }
                     }
                 }
@@ -773,6 +803,15 @@ int parse_args(int argc, char *argv[], char ** filename, chipset_type *forced_ch
         else if (!strcmp(argv[index], "945GM")) {
             *forced_chipset = CT_945GM;
         }
+        else if (!strcmp(argv[index], "946GZ")) {
+            *forced_chipset = CT_946GZ;
+        }
+        else if (!strcmp(argv[index], "G965")) {
+            *forced_chipset = CT_G965;
+        }
+        else if (!strcmp(argv[index], "Q965")) {
+            *forced_chipset = CT_Q965;
+        }
         else {
             *forced_chipset = CT_UNKWN;
         }
@@ -871,9 +910,13 @@ int main (int argc, char *argv[]) {
     }
 
     if (mode!=0 && x!=0 && y!=0) {
-        unlock_vbios(map);
+        if (!filename) 
+            unlock_vbios(map);
+
         set_mode(map, mode, x, y, bp, htotal, vtotal);
-        relock_vbios(map);
+
+        if (!filename)
+            relock_vbios(map);
         
         printf("Patch mode %02x to resolution %dx%d complete\n", mode, x, y);
         
